@@ -3,12 +3,18 @@ maniphono __init__.py
 """
 
 # Package information
-__version__ = "0.3.2"  # remember to sync with setup.py
+__version__ = "0.4.1"  # remember to sync with setup.py
 __author__ = "Tiago Tresoldi"
 __email__ = "tiago.tresoldi@lingfil.uu.se"
 
 # Import from the various modules
-from maniphono.phonomodel import PhonoModel, model_mipa, model_tresoldi
+from maniphono.phonomodel import (
+    HumanModel,
+    MachineModel,
+    model_mipa,
+    model_tresoldi,
+    model_encoder,
+)
 from maniphono.sound import Sound
 from maniphono.segment import BoundarySegment, SoundSegment, parse_segment
 from maniphono.segsequence import SegSequence, parse_sequence
@@ -18,18 +24,18 @@ from maniphono.common import (
     parse_constraints,
     parse_fvalues,
 )
-from maniphono.metrics import DistanceRegressor
 
 # Build the exported namespace; note that functions from
 # the common/utils module are not included (but they are available
 # with full qualified usage, like `maniphono.codepoint2glyph`)
 __all__ = [
-    "PhonoModel",
+    "HumanModel",
+    "MachineModel",
     "model_mipa",
     "model_tresoldi",
+    "model_encoder",
     "Sound",
     "SoundSegment",
     "SegSequence",
     "parse_sequence",
-    "DistanceRegressor",
 ]
